/*
   Swap Chaining Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to create an application that will render a 3D scene
   to two different windows using nothing more than two swap chain objects.
*/


// Demo: Swap Chaining.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN

#include<d3dx9.h>


// Function Prototypes...
bool InitializeDirect3D(bool fullscreen);
void RenderScene();
void RenderScene2();
void SetRotation();
void SetView();
void SetPerpective();
void ShutdownDirect3D();


// Here are our Direct3D objects.
LPDIRECT3D9 Direct3D_Object = NULL;
LPDIRECT3DDEVICE9 D3D_Device = NULL;

LPD3DXBUFFER MaterialBuffer = NULL;          // Material buffer.
LPD3DXMESH Teapot = NULL;                    // Mesh object.
D3DMATERIAL9 TeapotMaterial;                 // Objects material.

// This is the rotation angle.
float angle = 0.0f;

// Window handles to both windows that will open.
HWND g_hwnd = NULL;
HWND g_hwnd2 = NULL;

// D3D swap chain objects.  One for each window.
LPDIRECT3DSWAPCHAIN9 chain1 = NULL;
LPDIRECT3DSWAPCHAIN9 chain2 = NULL;

// Scene light source.
D3DLIGHT9 light;


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message)
       {
		   case WM_KEYUP:
            // If the user presses the escape key then exit the application.
            if(wParam == VK_ESCAPE)
               PostQuitMessage(0);
            break;

         case WM_DESTROY:
         case WM_CLOSE:
            PostQuitMessage(0);
            break;

            default:
               break;
       }

    return DefWindowProc( hwnd, message, wParam, lParam );
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                           // Message object.
   WNDCLASSEX windowClass;                            // Window class object.
   bool done = false;                                 // true if we should exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);           // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;       // style of the window.
   windowClass.lpfnWndProc = WndProc;                 // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                        // Extra class information.
   windowClass.cbWndExtra = 0;                        // Extra window information.
   windowClass.hInstance = hInstance;                 // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Window Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW); // Load mouse cursor.
   windowClass.hbrBackground = NULL;                  // Background color.
   windowClass.lpszMenuName = NULL;                   // Menu.
   windowClass.lpszClassName = "UGPClass";            // Name of the window class.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// Minimized window icon.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // After your class has been registered then you are ready to create your window.
   g_hwnd = CreateWindowEx(NULL,                                     // The extended style.
                         "UGPClass",                                 // Window class.
                         "Swap Chaining - by The Programming Ace",   // Window name.
                          WS_OVERLAPPEDWINDOW | WS_VISIBLE |         // Window style.
                          WS_SYSMENU |WS_CLIPCHILDREN |              // Window style.
                          WS_CLIPSIBLINGS,                           // Window style.
                          100, 100,                                  // X, Y coords.
                          640, 480,                                  // Window size.
                          NULL,                                      // Handle to parent window.
                          NULL,                                      // Menu.
                          hInstance,                                 // Handle to app instance.
                          NULL);                                     // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!g_hwnd) return 0;

   ShowWindow(g_hwnd, SW_SHOW);    // Show the window.
   UpdateWindow(g_hwnd);           // Update its display.

   g_hwnd2 = CreateWindowEx(NULL,                                    // The extended style.
                         "UGPClass",                                 // Window class.
                         "Swap Chaining window 2",                   // Window name.
                          WS_OVERLAPPEDWINDOW | WS_VISIBLE |         // Window style.
                          WS_SYSMENU |WS_CLIPCHILDREN |              // Window style.
                          WS_CLIPSIBLINGS,                           // Window style.
                          300, 300,                                  // X, Y coords.
                          640, 480,                                  // Window size.
                          NULL,                                      // Handle to parent window.
                          NULL,                                      // Menu.
                          hInstance,                                 // Handle to app instance.
                          NULL);                                     // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!g_hwnd2) return 0;

   ShowWindow(g_hwnd2, SW_SHOW);    // Show the window.
   UpdateWindow(g_hwnd2);           // Update its display.

   done = false;                    // false = run program, true means stop.

   // Initialize Direct3D.  If fails then we don't want to run the application.
   if(!InitializeDirect3D(false)) done = true;

   // Application loop.
   while(!done)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     done = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
               RenderScene2();
            }
      }

    // Here we shutdown Direct 3D.
	ShutdownDirect3D();

   // Here we unregister the window class with the OS.
   UnregisterClass("UGPClass", windowClass.hInstance);

   return (int)msg.wParam;
}


bool InitializeDirect3D(bool fullscreen)
{
   // This object will allow us to set the display mode of the screen.
   D3DDISPLAYMODE DisplayMode;

   // This will allow us to set the parameters of the screen.
   D3DPRESENT_PARAMETERS Present_Parameters;

   // This is used to get the capabilities of the hardware.
   D3DCAPS9 D3DCaps;

   // It is always a good idea to clear out memory in object although not necessary.
   ZeroMemory(&Present_Parameters, sizeof(Present_Parameters));

   // Create the Direct3D object to get everything started.
   Direct3D_Object = Direct3DCreate9(D3D_SDK_VERSION);

   // Error checking.  Make sure that it was successful.
   if(Direct3D_Object == NULL)
      {
         MessageBox(NULL, "Error, couldn't initialize DirectX!?!",
                    "Error!", MB_OK);
         return false;
      }

   // This function will get the display mode of the device and place it in DisplayMode.
   if(FAILED(Direct3D_Object->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &DisplayMode)))
      {
         MessageBox(NULL, "Error setting the display mode.", "Error!", MB_OK);
         return false;
      }

   // Get the capabilities of the hardware.
	if(FAILED(Direct3D_Object->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &D3DCaps)))
		return false;

   // Test which is supported, hardware or software vertex processing.
	DWORD VertexProcessing = 0;

	if(D3DCaps.VertexProcessingCaps != 0)
		VertexProcessing |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		VertexProcessing |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

   // Here we are setting the applications parameters...
   if(fullscreen)
      {
         Present_Parameters.Windowed = FALSE;               // Window mode (fullscreen).
         Present_Parameters.BackBufferWidth = DisplayMode.Width;
         Present_Parameters.BackBufferHeight = DisplayMode.Height;
      }
   else
      Present_Parameters.Windowed = TRUE;                   // Window mode (not fullscreen).
   Present_Parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;   // Dealing with animation (see doc).
   Present_Parameters.BackBufferFormat = DisplayMode.Format;// Render to the area of the screen.
   Present_Parameters.BackBufferCount = 1;                  // Number of back buffers.
   Present_Parameters.EnableAutoDepthStencil = TRUE;        // Check documentation.
   Present_Parameters.AutoDepthStencilFormat = D3DFMT_D16;  // Check documentation.

   // Now we must create the rendering device.
   if(FAILED(Direct3D_Object->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, g_hwnd,
                                           VertexProcessing,
                                           &Present_Parameters, &D3D_Device)))
      {
         MessageBox(NULL, "CreateDevice() failed!  Make sure you have DirectX 9.",
                    "Error!", MB_OK);
         return false;
      }

   // One last check to be sure.
   if(D3D_Device == NULL)
      {
         MessageBox(NULL, "D3D_Device is equal to NULL!?!", "Error!", MB_OK);
         return false;
      }

   // Setup the light source.
   light.Type = D3DLIGHT_DIRECTIONAL;
   light.Direction = D3DXVECTOR3(0.0f, 0.0f, 1.0f);

   light.Diffuse.r = light.Diffuse.g = light.Diffuse.b = light.Diffuse.a = 1.0f;
   light.Specular.r = light.Specular.g = light.Specular.b = light.Specular.a = 1.0f;

   D3D_Device->SetLight(0, &light);
   D3D_Device->LightEnable(0, TRUE);

   // Setup the material properties for the teapot.
   ZeroMemory(&TeapotMaterial, sizeof(D3DMATERIAL9));
   TeapotMaterial.Diffuse.r = TeapotMaterial.Ambient.r = 0.6f;
   TeapotMaterial.Diffuse.g = TeapotMaterial.Ambient.g = 0.6f;
   TeapotMaterial.Diffuse.b = TeapotMaterial.Ambient.b = 0.7f;
   TeapotMaterial.Specular.r = 0.4f;
   TeapotMaterial.Specular.g = 0.4f;
   TeapotMaterial.Specular.b = 0.4f;
   TeapotMaterial.Power = 8.0f;

   // First is for ambient lighting.
   D3D_Device->SetRenderState(D3DRS_AMBIENT, D3DCOLOR_COLORVALUE(0.3f, 0.3f, 0.3f, 1.0f));
	D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);// Culling off.
   D3D_Device->SetRenderState(D3DRS_ZENABLE, TRUE);         // Enable depth testing.

   // To create the teapot we simple call D3DXCreateTeapot and the rest is done for us.
   if(FAILED(D3DXCreateTeapot(D3D_Device, &Teapot, &MaterialBuffer)))
      return false;


   // We create the first chain by simple calling GetSwapChain to get the default.
   // Then we call CreateAdditionalSwapChain to create the second window's chain.
   D3D_Device->GetSwapChain(0, &chain1);
   D3D_Device->CreateAdditionalSwapChain(&Present_Parameters, &chain2);


   // Set the window perpective.
   SetPerpective();

   return true;
}


void SetPerpective()
{
   // This matrix will do nothing more but tell Direct3D how far back it can see
   // and how close it can see.
   D3DXMATRIXA16 Projection;

   // Set the perspective information.
   D3DXMatrixPerspectiveFovLH(&Projection, 45.0f, 640/480, 0.1f, 500.0f);

   // Set the projection.
   D3D_Device->SetTransform(D3DTS_PROJECTION, &Projection);
}


void SetRotation()
{
   // This will hold the world matrix.
   D3DXMATRIXA16 World;

   // Set the rotation value to the matrix.
   D3DXMatrixRotationY(&World, angle);

   // Set the matrix.
   D3D_Device->SetTransform(D3DTS_WORLD, &World);

   // Add 0.01 to angle to rotate the triangle even more the next frame.
   angle += 0.01f;

   if(angle >= 360) angle = 0.0f;
}


void SetView()
{
   // This will hold the view matrix.
   D3DXMATRIXA16 View;

   D3DXVECTOR3 Eye(0.0f, 0.0f, -5.0f);     // Position of view.
   D3DXVECTOR3 LookAt(0.0f, 0.0f, 1.0f);   // Point view is looking at.
   D3DXVECTOR3 Up(0.0f, 1.0f, 0.0f);       // Which way is up.

   // Set all the values this matrix will need to know (position, look at point, etc).
   D3DXMatrixLookAtLH(&View, &Eye, &LookAt, &Up);

   // Move the "view" to the position and values we set above.
   D3D_Device->SetTransform(D3DTS_VIEW, &View);
}


void RenderScene()
{
   LPDIRECT3DSURFACE9 pBackBuffer = NULL;

	chain1->GetBackBuffer(0, D3DBACKBUFFER_TYPE_MONO, &pBackBuffer);
	D3D_Device->SetRenderTarget(0, pBackBuffer);

   // Clear the screen to black.
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_XRGB(0,0,0), 1.0f, 0);
    
   // Tell Direct 3D we are now going to start drawing.
   D3D_Device->BeginScene();

      SetRotation();// Set the world matrix so we can rotate.
      SetView();    // Set the view matrix (view position).

      // Set the material and draw the Teapot.
      D3D_Device->SetMaterial(&TeapotMaterial);
      Teapot->DrawSubset(0);

   // Done drawing.
   D3D_Device->EndScene();
    
   // Swap the old frame with the new one.
   chain1->Present(NULL, NULL, g_hwnd, NULL, 0);
}


void RenderScene2()
{
   LPDIRECT3DSURFACE9 pBackBuffer = NULL;

	chain2->GetBackBuffer(0, D3DBACKBUFFER_TYPE_MONO, &pBackBuffer);
	D3D_Device->SetRenderTarget(0, pBackBuffer);

   // Clear the screen to black.
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_XRGB(0,0,0), 1.0f, 0);
    
   // Tell Direct 3D we are now going to start drawing.
   D3D_Device->BeginScene();

      SetRotation();// Set the world matrix so we can rotate.
      SetView();    // Set the view matrix (view position).

      // Set the material and draw the teapot.
      D3D_Device->SetMaterial(&TeapotMaterial);
      Teapot->DrawSubset(0);

   // Done drawing.
   D3D_Device->EndScene();
    
   // Swap the old frame with the new one.
   chain2->Present(NULL, NULL, g_hwnd2, NULL, 0);
}


void ShutdownDirect3D()
{
   // Here we release the Direct3D objects.
   if(D3D_Device != NULL)
      {
         D3D_Device->Release();
         D3D_Device = NULL;
      }

   if(Direct3D_Object != NULL)
      {
         Direct3D_Object->Release();
         Direct3D_Object = NULL;
      }
   
   if(MaterialBuffer != NULL)
      {
         MaterialBuffer->Release();
         MaterialBuffer = NULL;
      }

   if(Teapot != NULL)
      {
         Teapot->Release();
         Teapot = NULL;
      }

   if(chain1 != NULL)
      {
         chain1->Release();
         chain2 = NULL;
      }

   if(chain2 != NULL)
      {
         chain2->Release();
         chain2 = NULL;
      }
}


// Recap:

// Using swap chains can allow you to render to two different windows with the
// same device.  This is simple to do in DirectX using swap chain objects.  This
// is also simple to do in OpenGL although OpenGL don't have swap chain objects.


// Copyright September 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com